/*------------------------------------------------------------------------------
* lambda.c : lambda/mlambda inteter least square estimation
*
* references
* (1) P.J.G.Teunissen, The least-square ambiguity decorrelation adjustment: a
*     method for fast GPS ambiguity estimation, J.Geodesy, Vol.70, 65-82, 1995
* (2) X.-W.Chang, X.Yang, T.Zhou, MLAMBDA: A modified LAMBDA method for integer
*     least-squares estimation, J.Geodesy, Vol.79, 552-565, 2005
*
*          Copyright (C) 2006 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.1 $ $Date: 2006/07/12 03:59:22 $
* history : 2006/07/11 0.1 new
*           2006/12/01 0.2 support mkl, fix bug of initialization of Z
*-----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "lambda.h"
#ifdef MKL
#include <mkl_types.h>
#include <mkl_blas.h>
#include <mkl_lapack.h>
#endif
#define LOOPMAX     10000           /* maximum count of search loop */
#define SQR(x)      ((x)*(x))
#define SGN(x)      ((x)<=0.0?-1.0:1.0)
#define ROUND(x)    (floor((x)+0.5))
#define SWAP(x,y)   {double tmp_; tmp_=x; x=y; y=tmp_;}
#define MAT(n)      (double *)malloc(sizeof(double)*(n))

/* LAPACK/BLAS proto-types ---------------------------------------------------*/
#ifdef MKL
#define dgemv_		degemv
#define dgetrf_		dgetrf
#define dgetrs_		dgetrs
#endif
#ifdef __cplusplus
extern "C" {
#endif
extern void dgemv_(char *trans, int *m, int *n, double *alpha, double *a,
                   int *lda, double *x, int *incx, double *beta, double *y,
                   int *incy);
extern void dgetrf_(int *m, int *n, double *a, int *lda, int *ipiv, int *info);
extern void dgetrs_(char *trans, int *n, int *m, double *a, int *lda, int *ipiv,
                    double *b, int *ldb, int *info);
#ifdef __cplusplus
}
#endif

/* LD factorization : Q=L'*diag(D)*L -----------------------------------------*/
static int LD(int n, const double *Q, double *L, double *D)
{
    int i,j,k;
    double a,*A=MAT(n*n);
    
    if (!A) {
        fprintf(stderr,"%s : malloc error\n",__FILE__);
        return -9;
    }
    memcpy(A,Q,sizeof(double)*n*n);
    for (i=0;i<n*n;i++) L[i]=0.0;
    for (i=n-1;i>=0;i--) {
        if ((D[i]=A[i+n*i])<=0.0) {
            fprintf(stderr,"%s : LD factrorization error\n",__FILE__);
            free(A);
            return i+1;
        }
        a=sqrt(D[i]);
        for (j=0;j<=i;j++) L[i+n*j]=A[i+n*j]/a;
        for (j=0;j<=i-1;j++) {
            for (k=0;k<=j;k++) A[j+n*k]-=L[i+n*k]*L[i+n*j];
        }
        for (j=0;j<=i;j++) L[i+n*j]/=L[i+n*i];
    }
    free(A);
    return 0;
}
/* integer gauss transformation ----------------------------------------------*/
static void gauss(int n, double *L, double *Z, int i, int j)
{
    int k,mu;
    
    if ((mu=(int)ROUND(L[i+n*j]))!=0) {
        for (k=i;k<n;k++) L[k+n*j]-=(double)mu*L[k+n*i];
        for (k=0;k<n;k++) Z[k+n*j]-=(double)mu*Z[k+n*i];
    }
}
/* permutations --------------------------------------------------------------*/
static void perm(int n, double *L, double *D, int j, double del, double *Z)
{
    int k;
    double eta,lam,a0,a1;
    
    eta=D[j]/del;
    lam=D[j+1]*L[j+1+n*j]/del;
    D[j]=eta*D[j+1];
    D[j+1]=del;
    for (k=0;k<=j-1;k++) {
        a0=L[j+n*k];
        a1=L[j+1+n*k];
        L[j+n*k]=-L[j+1+n*j]*a0+a1;
        L[j+1+n*k]=eta*a0+lam*a1;
    }
    L[j+1+n*j]=lam;
    for (k=j+2;k<n;k++) SWAP(L[k+n*j],L[k+n*(j+1)]);
    for (k=0;k<n;k++) SWAP(Z[k+n*j],Z[k+n*(j+1)]);
}
/* reduction (lambda decorrelation) : z=Z'*a, Qz=Z'*Q*Z=L'*diag(D)*L ---------*/
static void reduction(int n, double *L, double *D, double *Z)
{
    int i,j,k;
    double del;
    
    for (i=0;i<n*n;i++) Z[i]=0.0;
    for (i=0;i<n;i++) Z[i+n*i]=1.0;
    j=n-2; k=n-2;
    while (j>=0) {
        if (j<=k) {
            for (i=j+1;i<n;i++) gauss(n,L,Z,i,j);
        }
        del=D[j]+SQR(L[j+1+n*j])*D[j+1];
        if (del<D[j+1]) {
            perm(n,L,D,j,del,Z);
            k=j; j=n-2;
        }
        else j--;
    }
}
/* modified lambda search with shrinking ------------------------------------*/
static int search(int n, int m, const double *L, const double *D,
                  const double *zs, double *zn, double *s)
{
    int i,j,k,c,nn=0,imax=0;
    double newdist,maxdist=1E99,y;
    double *S=MAT(n*n),*dist=MAT(n),*zb=MAT(n),*z=MAT(n),*step=MAT(n);
    
    if (!S||!dist||!zb||!z||!step) {
        fprintf(stderr,"%s : malloc error\n",__FILE__);
        free(S); free(dist); free(zb); free(z); free(step);
        return -9;
    }
    k=n-1; dist[k]=0.0;
    zb[k]=zs[k];
    z[k]=ROUND(zb[k]); y=zb[k]-z[k]; step[k]=SGN(y);
    for (i=0;i<n*n;i++) S[i]=0.0;
    for (c=0;c<LOOPMAX;c++) {
        newdist=dist[k]+SQR(y)/D[k];
        if (newdist<maxdist) {
            if (k!=0) {
                dist[--k]=newdist;
                for (i=0;i<=k;i++)
                    S[k+n*i]=S[k+1+n*i]+(z[k+1]-zb[k+1])*L[k+1+n*i];
                zb[k]=zs[k]+S[k+n*k];
                z[k]=ROUND(zb[k]); y=zb[k]-z[k]; step[k]=SGN(y);
            }
            else {
                if (nn<m) {
                    if (nn==0||newdist>s[imax]) imax=nn;
                    for (i=0;i<n;i++) zn[i+n*nn]=z[i];
                    s[nn++]=newdist;
                }
                else {
                    if (newdist<s[imax]) {
                        for (i=0;i<n;i++) zn[i+n*imax]=z[i];
                        s[imax]=newdist;
                        for (i=imax=0;i<m;i++) if (s[imax]<s[i]) imax=i;
                    }
                    maxdist=s[imax];
                }
                z[0]+=step[0]; y=zb[0]-z[0]; step[0]=-step[0]-SGN(step[0]);
            }
        }
        else {
            if (k==n-1) break;
            else {
                k++;
                z[k]+=step[k]; y=zb[k]-z[k]; step[k]=-step[k]-SGN(step[k]);
            }
        }
    }
    for (i=0;i<m-1;i++) { /* sort by s */
        for (j=i+1;j<m;j++) {
            if (s[i]<s[j]) continue;
            SWAP(s[i],s[j]);
            for (k=0;k<n;k++) SWAP(zn[k+n*i],zn[k+n*j]);
        }
    }
    free(S); free(dist); free(zb); free(z); free(step);
    
    if (c>=LOOPMAX) {
        fprintf(stderr,"%s : search loop count overflow\n",__FILE__);
        return -1;
    }
    return 0;
}
/* lambda/mlambda integer least-square estimation ------------------------------
* args   : int    n      I  number of estimated float ambiguities
*          int    m      I  number of fix solutions
*          double *a     I  float ambiguities (n x 1)
*          double *Q     I  covariance matrix of float ambiguities (n x n)
*          double *F     O  fixed solutions (n x m)
*          double *s     O  sum of squared residulas of fixed solutions (1 x m)
* return : status (0:ok,other:error)
* notes  : matrix stored by column-major order (fortran convension)
*-----------------------------------------------------------------------------*/
extern int lambda(int n, int m, const double *a, const double *Q, double *F,
                  double *s)
{
    int stat,inc=1,*ipiv=(int *)malloc(sizeof(int)*n);
    double zero=0.0,one=1.0,*L=MAT(n*n),*D=MAT(n),*Z=MAT(n*n),*z=MAT(n);
    
    if (!L||!D||!Z||!z||!ipiv) {
        fprintf(stderr,"%s : malloc error\n",__FILE__);
        free(L); free(D); free(Z); free(z); free(ipiv); 
        return -9;
    }
    if (!(stat=LD(n,Q,L,D))) {
        reduction(n,L,D,Z);
        dgemv_("T",&n,&n,&one,Z,&n,(double *)a,&inc,&zero,z,&inc); /* z=Z'*a */
        if (!(stat=search(n,m,L,D,z,F,s))) {
            dgetrf_(&n,&n,Z,&n,ipiv,&stat); /* F=Z'\F */
            if (!stat) dgetrs_("T",&n,&m,Z,&n,ipiv,F,&n,&stat);
            if (stat) fprintf(stderr,"%s : back-subst error\n",__FILE__);
        }
    }
    free(L); free(D); free(Z); free(z); free(ipiv); 
    return stat;
}
