/*------------------------------------------------------------------------------
* rtkpos.c : rtk positioning
*
*          Copyright (C) 2006 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.2 $ $Date: 2006/12/12 23:27:39 $
* history : 2006/11/28 0.1 new
*-----------------------------------------------------------------------------*/
#include <string.h>
#include "rtkcmn.h"
#include "lambda.h"
#include "pntpos.h"
#include "rtkpos.h"

#define NF			1				/* number of carrier frequencies */
#define NR			3				/* number of real parameters */
#define NX			(NR+MAXSAT*NF)	/* number of all parameters */
#define SIGP		10.0			/* initial rover position std-dev (m) */
#define SIGB		30.0			/* initial phase-bias std-dev (cycle) */
#define SIGY		0.003			/* phase measurement error std-dev (m) */
#define MAXOUT		10				/* max data outage epochs to reset bias */
#define MAXINNO		30.0			/* threshold of innovation (m) */

#define IX(s)		(NR-1+(s))		/* state index of phase-bias */

static const double lam[]={C/FREQ1,C/FREQ2}; /* carrier wave length (m) */

/* single-difference of observable (k=0:L1,1:L2,2:C1,3:P2) -------------------*/
static double sdobs(const obsd_t *obs, int i, int j, int k)
{
	return k<2?lam[k]*(obs[i].L[k]-obs[j].L[k]):obs[i].P[k-2]-obs[j].P[k-2];
}
/* double-difference of observable -------------------------------------------*/
static double ddobs(const obsd_t *obs, const int *u, const int *r, int i,
					int j, int k)
{
	return sdobs(obs,u[i],r[i],k)-sdobs(obs,u[j],r[j],k);
}
/* initialize a state and its covariance -------------------------------------*/
static void initx(double *x, double *P, int n, double xi, double sig, int i)
{
	int j;
	x[i]=xi; for (j=0;j<n;j++) P[i+j*n]=P[j+i*n]=i==j?SQR(sig):0.0;
}
/* select common satellites --------------------------------------------------*/
static int selsats(const obsd_t *obs, const double *azel, int mu, int mr,
				   double elmin, int *s, int *u, int *r)
{
	int i,j,k=0;
	
	for (i=0,j=mu;i<mu&&j<mu+mr;i++,j++) {
		if		(obs[i].sat<obs[j].sat) j--;
		else if (obs[i].sat>obs[j].sat) i--;
		else if (azel[1+i*2]>=elmin*D2R&&!ISNAN(sdobs(obs,i,j,0))) {
			s[k]=obs[i].sat; u[k]=i; r[k++]=j;
		}
	}
	return k;
}
/* update phase-biases -------------------------------------------------------*/
static void udbias(double *x, double *P, int n, const obsd_t *obs, const int *s,
				   const int *u, const int *r, int m, int *outcnt)
{
	int i;
	double bias,offset=0.0;
	
	/* reset phase-bias when data outage counter expired */
	for (i=1;i<=MAXSAT;i++) {
		if (++outcnt[i-1]>MAXOUT&&x[IX(i)]!=0.0) initx(x,P,n,0.0,0.0,IX(i));
	}
	for (i=0;i<m;i++) {
		bias=(sdobs(obs,u[i],r[i],0)-sdobs(obs,u[i],r[i],2))/lam[0];
		
		/* initialize phase-bias at first epoch or slip */
		if (x[IX(s[i])]==0.0||obs[u[i]].LLI[0]||obs[r[i]].LLI[0])
			initx(x,P,n,bias,SIGB,IX(s[i]));
		
		offset+=bias-x[IX(s[i])];
	}
	/* add constraint to phase biases */
	for (i=1;i<=MAXSAT;i++) if (x[IX(i)]!=0.0) x[IX(i)]+=offset/m;
}
/* relative precise positioning ----------------------------------------------*/
static int precpos(double *x, double *P, int n, const obsd_t *obs, int mu,
				   int mr, const double *rs, const double *azel, int *outcnt,
				   const double *rb, double elmin)
{
	int i,j,k,m,s[MAXSAT],u[MAXSAT],r[MAXSAT],info=-9;
	double d[MAXSAT],eu[3*MAXSAT],er[3],h,v[MAXSAT],*H,*R=NULL;
	
	/* select common satellites */
	if ((m=selsats(obs,azel,mu,mr,elmin,s,u,r))<2) return -1;
	
	/* update phase-bias */
	udbias(x,P,n,obs,s,u,r,m,outcnt);
	
	/* single-difference geometric distance */
	for (i=0;i<m;i++) d[i]=geodist(rs+u[i]*3,x,eu+i*3)-geodist(rs+r[i]*3,rb,er);
	
	if (!(H=MAT(n*MAXSAT))) return -9;
	
	for (i=1,k=0;i<m;i++) {
		/* double difference phase model */
		h=d[i]-d[0]+lam[0]*(x[IX(s[i])]-x[IX(s[0])]);
		
		/* partial derivatives */
		for (j=0;j<n;j++) H[j+k*n]=j<3?-eu[j+i*3]+eu[j]:0.0;
		H[IX(s[i])+k*n]=lam[0]; H[IX(s[0])+k*n]=-lam[0];
		
		/* innovation */
		if (fabs(v[k]=ddobs(obs,u,r,i,0,0)-h)<=MAXINNO) {
			k++; outcnt[s[i]-1]=outcnt[s[0]-1]=0;
		}
	}
	if (k>0&&(R=MAT(k*k))) {
		/* covariances of measurement errors */
		for (i=0;i<k;i++) {
			for (j=0;j<k;j++) R[i+j*k]=R[j+i*k]=(i==j?4.0:2.0)*SQR(SIGY);
		}
		/* measurement update */
		info=filter(x,P,H,v,R,n,k);
	}
	free(H); free(R);
	return info;
}
/* transform to double-difference form ---------------------------------------*/
static void transdd(const double *x, const double *P, int n, const int *ix,
					int m, int r, double *b, double *Qb, double *Qr)
{
	int i,j,k=r+m; /* m = # of integer params, r = # of real params. */
	double *y=MAT(k),*A=MAT(k*n),*Q=MAT(k*k);
	
	if (!y||!A||!Q) {free(y); free(A); free(Q); return;}
	
	/* transform to double-difference states */
	for (i=0;i<k*n;i++) A[i]=0.0;
	for (i=0;i<r;i++) A[i+i*k]=1.0;
	for (i=r;i<k;i++) {A[i+ix[0]*k]=1.0; A[i+ix[i-r+1]*k]=-1.0;}
	transx(x,P,A,n,k,y,Q);
	for (i=0;i<m;i++) b[i]=y[r+i];
	
	/* covariances of ambiguities, real params - ambiguities */
	for (i=0;i<m;i++) for (j=0;j<m;j++) Qb[i+j*m]=Q[r+i+(r+j)*k];
	for (i=0;i<m;i++) for (j=0;j<r;j++) Qr[i+j*m]=Q[r+i+j*k];
	free(y); free(A); free(Q);
}
/* integer ambiguity validation ----------------------------------------------*/
static int valamb(const double *s)
{
	/* popular ratio test */
	return s[1]/s[0]>=3.0?0:-1;
}
/* integer ambiguity resolution ----------------------------------------------*/
static int fixamb(const double *x, const double *P, int n, double *rr)
{
	int i,j=1,k=NR,m,info,ix[MAXSAT],ipiv[MAXSAT];
	double *b,*Qb,*Qr,*f,s[2],one=1.0,min=-1.0;
	
	for (i=0;i<NR;i++) rr[i]=x[i];
	
	/* valid phase-bias indexes */
	for (i=1,m=0;i<=MAXSAT;i++) if(x[IX(i)]!=0.0) ix[m++]=IX(i);
	if (m>0) m-=NR-1; else return 0;
	
	b=MAT(m); Qb=MAT(m*m); Qr=MAT(m*NR); f=MAT(m*2);
	if (!b||!Qb||!Qr||!f) {free(b); free(Qb); free(Qr); free(f); return 1;}
	
	/* transform to double-difference form */
	transdd(x,P,n,ix,m,NR,b,Qb,Qr);

	/* lambda/mlambda integer least square estimation */
	if (!(info=lambda(m,2,b,Qb,f,s))) {
		
		/* integer ambiguity validation */
		if (!(info=valamb(s))) {
			
			/* compute fixed receiver postion : rr=rr-Qr*Qb\(b-f) */
			for (i=0;i<m;i++) b[i]-=f[i];
			dgetrf_(&m,&m,Qb,&m,ipiv,&info);
			if (!info) dgetrs_("N",&m,&j,Qb,&m,ipiv,b,&m,&info);
			if (!info) dgemm_("T","N",&k,&j,&m,&min,Qr,&m,b,&m,&one,rr,&k);
		}
	}
	free(b); free(Qb); free(Qr); free(f);
	return !info?2:1;
}
/* rtk positioning -------------------------------------------------------------
* input obs/nav data, compute receiver positions and output result
* args   : infunc_t input 	I 	data input callback
*		   outfunc_t output	I	data output callback
*		   int    mode		I	positioning mode (0:static,1:kinematic)
*		   double *rb		I	reference station position (ecef) (m)
*		   double elmin		I	elevation cutoff angle (deg)
* return : none
* note   : data input and output callback shall comply the following prototypes.
*          int  func(obsd_t *obs, nav_t *nav);
*          void func(gtime_t t, double *rr);
*-----------------------------------------------------------------------------*/
extern void rtkpos(infunc_t input, outfunc_t output, int mode, const double *rb,
				   double elmin)
{
	int i,n,m,outcnt[MAXSAT]={0};
	double x[NX]={0.0},P[NX*NX]={0.0},rr[7],dtr[2];
	double rs[3*MAXOBS],dts[MAXOBS],azel[2*MAXOBS];
	obsd_t *obs;
	nav_t *nav;
	
	trace_("rtkpos()\n");
	
	while ((n=(*input)(&obs,&nav))>0) { /* input callback */
		
		for (m=0;m<n&&obs[m].rcv==1;m++) continue;
		
		/* satellite position/clock-bias */
		satpos(obs,n,nav,rs,dts);
		
		/* single point positioning */
		if (pointp(obs,m,rs,dts,nav->ion,elmin,rr,dtr,azel)) continue;
		
		/* initialize rover position */
		if (mode||NORM(x)<=0.0) {
			for (i=0;i<3;i++) initx(x,P,NX,rr[i],SIGP,i);
		}
		/* precise relative positioning */
		if (!precpos(x,P,NX,obs,m,n-m,rs,azel,outcnt,rb,elmin)) {
			
			/* integer ambiguity resolution */
			rr[6]=fixamb(x,P,NX,rr);
			
			for (i=0;i<3;i++) rr[3+i]=sqrt(P[i+i*NX]);
		}
		else for (i=3;i<7;i++) rr[i]=0.0;
		
		(*output)(timeadd(obs[0].time,-dtr[0]),rr); /* output callback */
	}
}
